# contributed by Luke Olson

import numpy as np

n = 50
x, y = np.meshgrid(np.linspace(-1, 1, n),
np.linspace(-1, 1, n))

u = np.exp(-50 * (x**2 + y**2))

from pyvisfile.vtk import write_structured_grid
mesh = np.rollaxis(np.dstack((x, y)), 2)
write_structured_grid("test.vts", mesh,
        point_data=[("u", u[np.newaxis, :, :])])



__copyright__ = "Copyright (C) 2007,2010 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

__doc__ = """PyVisfile exposes the functionality of libsilo to Python using the
Boost.Python wrapper library.

To use pyvisfile, you would typically create a SiloFile instance and then
write different entities (variables and meshes, for the most part) to
this file.

If you are running on a parallel machine, you might want to use
ParallelSiloFile to automatically create a master file along with your
SiloFile.
"""

try:
    import pyvisfile.silo._internal as _silo
except ImportError:
    from warnings import warn
    warn("Importing the native-code parts of PyVisfile's silo component failed. "
            "By default, PyVisfile is installed without Silo support. If you would "
            "like support for the Silo file format, configure with --use-silo. "
            "This requires the libsilo library.")
    raise


from pyvisfile.silo._internal import (  # noqa: F401
        get_silo_version, set_deprecate_warnings,
        # enums
        DBObjectType, DBdatatype,
        # classes
        DBToc, DBCurve, DBQuadMesh, DBQuadVar, IntVector,
        )

from pyvisfile.silo._internal import (  # noqa: F401
        DB_NETCDF, DB_PDB, DB_TAURUS, DB_UNKNOWN, DB_DEBUG, DB_HDF5X,
        DB_PDBP, DB_HDF5,
        )
from pyvisfile.silo._internal import DB_CLOBBER, DB_NOCLOBBER   # noqa: F401
from pyvisfile.silo._internal import DB_READ, DB_APPEND         # noqa: F401
from pyvisfile.silo._internal import (  # noqa: F401
        DB_LOCAL, DB_SUN3, DB_SUN4, DB_SGI, DB_RS6000, DB_CRAY, DB_INTEL,
        )
from pyvisfile.silo._internal import (  # noqa: F401
        DBOPT_ALIGN, DBOPT_COORDSYS, DBOPT_CYCLE, DBOPT_FACETYPE, DBOPT_HI_OFFSET,
        DBOPT_LO_OFFSET, DBOPT_LABEL, DBOPT_XLABEL, DBOPT_YLABEL, DBOPT_ZLABEL,
        DBOPT_MAJORORDER, DBOPT_NSPACE, DBOPT_ORIGIN, DBOPT_PLANAR, DBOPT_TIME,
        DBOPT_UNITS, DBOPT_XUNITS, DBOPT_YUNITS, DBOPT_ZUNITS, DBOPT_DTIME,
        DBOPT_USESPECMF, DBOPT_XVARNAME, DBOPT_YVARNAME, DBOPT_ZVARNAME,
        DBOPT_ASCII_LABEL, DBOPT_MATNOS, DBOPT_NMATNOS, DBOPT_MATNAME, DBOPT_NMAT,
        DBOPT_NMATSPEC, DBOPT_BASEINDEX, DBOPT_ZONENUM, DBOPT_NODENUM,
        DBOPT_BLOCKORIGIN, DBOPT_GROUPNUM, DBOPT_GROUPORIGIN, DBOPT_NGROUPS,
        DBOPT_MATNAMES, DBOPT_EXTENTS_SIZE, DBOPT_EXTENTS, DBOPT_MATCOUNTS,
        DBOPT_MATLISTS, DBOPT_MIXLENS, DBOPT_ZONECOUNTS, DBOPT_HAS_EXTERNAL_ZONES,
        DBOPT_PHZONELIST, DBOPT_MATCOLORS, DBOPT_BNDNAMES, DBOPT_REGNAMES,
        DBOPT_ZONENAMES, DBOPT_HIDE_FROM_GUI, DBOPT_TOPO_DIM, DBOPT_REFERENCE,
        DBOPT_GROUPINGS_SIZE, DBOPT_GROUPINGS, DBOPT_GROUPINGNAMES, DBOPT_ALLOWMAT0,
        DBOPT_MRGTREE_NAME, DBOPT_REGION_PNAMES, DBOPT_TENSOR_RANK, DBOPT_MMESH_NAME,
        DBOPT_TV_CONNECTIVITY, DBOPT_DISJOINT_MODE, DBOPT_MRGV_ONAMES,
        DBOPT_MRGV_RNAMES, DBOPT_SPECNAMES, DBOPT_SPECCOLORS, DBOPT_LLONGNZNUM,
        DBOPT_CONSERVED, DBOPT_EXTENSIVE, DBOPT_MB_FILE_NS, DBOPT_MB_BLOCK_NS,
        DBOPT_MB_BLOCK_TYPE, DBOPT_MB_EMPTY_LIST, DBOPT_MB_EMPTY_COUNT,
        DBOPT_MB_REPR_BLOCK_IDX, DBOPT_MISSING_VALUE, DBOPT_ALT_ZONENUM_VARS,
        DBOPT_ALT_NODENUM_VARS, DBOPT_GHOST_NODE_LABELS, DBOPT_GHOST_ZONE_LABELS,
        )
from pyvisfile.silo._internal import (  # noqa: F401
        DB_TOP, DB_NONE, DB_ALL, DB_ABORT, DB_SUSPEND, DB_RESUME, DB_ALL_AND_DRVR,
        )
from pyvisfile.silo._internal import (  # noqa: F401
        E_NOERROR, E_BADFTYPE, E_NOTIMP, E_NOFILE, E_INTERNAL, E_NOMEM, E_BADARGS,
        E_CALLFAIL, E_NOTFOUND, E_TAURSTATE, E_MSERVER, E_PROTO, E_NOTDIR,
        E_MAXOPEN, E_NOTFILTER, E_MAXFILTERS, E_FEXIST, E_FILEISDIR, E_FILENOREAD,
        E_SYSTEMERR, E_FILENOWRITE, E_INVALIDNAME, E_NOOVERWRITE, E_CHECKSUM,
        E_COMPRESSION, E_GRABBED, E_NOTREG, E_CONCURRENT, E_DRVRCANTOPEN,
        E_BADOPTCLASS, E_NOTENABLEDINBUILD, E_MAXFILEOPTSETS, E_NOHDF5,
        E_EMPTYOBJECT, E_OBJBUFFULL,
        )
from pyvisfile.silo._internal import DB_ROWMAJOR, DB_COLMAJOR       # noqa: F401
from pyvisfile.silo._internal import (  # noqa: F401
        DB_COLLINEAR, DB_NONCOLLINEAR, DB_QUAD_RECT, DB_QUAD_CURV,
        )
from pyvisfile.silo._internal import (  # noqa: F401
        DB_NOTCENT, DB_NODECENT, DB_ZONECENT, DB_FACECENT, DB_BNDCENT,
        DB_EDGECENT, DB_BLOCKCENT,
        )
from pyvisfile.silo._internal import (  # noqa: F401
        DB_CARTESIAN, DB_CYLINDRICAL, DB_SPHERICAL, DB_NUMERICAL, DB_OTHER,
        )
from pyvisfile.silo._internal import DB_RECTILINEAR, DB_CURVILINEAR  # noqa: F401
from pyvisfile.silo._internal import DB_AREA, DB_VOLUME             # noqa: F401
from pyvisfile.silo._internal import DB_ON, DB_OFF                  # noqa: F401
from pyvisfile.silo._internal import DB_ABUTTING, DB_FLOATING       # noqa: F401
from pyvisfile.silo._internal import (  # noqa: F401
        DB_VARTYPE_SCALAR, DB_VARTYPE_VECTOR, DB_VARTYPE_TENSOR,
        DB_VARTYPE_SYMTENSOR, DB_VARTYPE_ARRAY, DB_VARTYPE_MATERIAL,
        DB_VARTYPE_SPECIES, DB_VARTYPE_LABEL,
        )
from pyvisfile.silo._internal import (  # noqa: F401
        DB_GHOSTTYPE_NOGHOST, DB_GHOSTTYPE_INTDUP,
        )
from pyvisfile.silo._internal import (  # noqa: F401
        DBCSG_QUADRIC_G, DBCSG_SPHERE_PR, DBCSG_ELLIPSOID_PRRR, DBCSG_PLANE_G,
        DBCSG_PLANE_X, DBCSG_PLANE_Y, DBCSG_PLANE_Z, DBCSG_PLANE_PN,
        DBCSG_PLANE_PPP, DBCSG_CYLINDER_PNLR, DBCSG_CYLINDER_PPR,
        DBCSG_BOX_XYZXYZ, DBCSG_CONE_PNLA, DBCSG_CONE_PPA, DBCSG_POLYHEDRON_KF,
        DBCSG_HEX_6F, DBCSG_TET_4F, DBCSG_PYRAMID_5F, DBCSG_PRISM_5F,
        )
from pyvisfile.silo._internal import (  # noqa: F401
        DBCSG_QUADRATIC_G, DBCSG_CIRCLE_PR, DBCSG_ELLIPSE_PRR, DBCSG_LINE_G,
        DBCSG_LINE_X, DBCSG_LINE_Y, DBCSG_LINE_PN, DBCSG_LINE_PP, DBCSG_BOX_XYXY,
        DBCSG_ANGLE_PNLA, DBCSG_ANGLE_PPA, DBCSG_POLYGON_KP, DBCSG_TRI_3P,
        DBCSG_QUAD_4P,
        )
from pyvisfile.silo._internal import (  # noqa: F401
        DBCSG_INNER, DBCSG_OUTER, DBCSG_ON, DBCSG_UNION, DBCSG_INTERSECT,
        DBCSG_DIFF, DBCSG_COMPLIMENT, DBCSG_XFORM, DBCSG_SWEEP,
        )

if get_silo_version() >= (4, 6, 1):
    from pyvisfile.silo._internal import (  # noqa: F401
            DB_HDF5_SEC2, DB_HDF5_STDIO, DB_HDF5_CORE, DB_HDF5_LOG,
            DB_HDF5_SPLIT, DB_HDF5_DIRECT, DB_HDF5_FAMILY, DB_HDF5_MPIO,
            DB_HDF5_MPIOP, DB_HDF5_MPIP, DB_HDF5_SILO,
            )
    from pyvisfile.silo._internal import (  # noqa: F401
            DB_ZONETYPE_BEAM, DB_ZONETYPE_TRIANGLE, DB_ZONETYPE_QUAD,
            DB_ZONETYPE_POLYHEDRON, DB_ZONETYPE_TET, DB_ZONETYPE_PYRAMID,
            DB_ZONETYPE_PRISM, DB_ZONETYPE_HEX,
            )


def _convert_optlist(ol_dict):
    optcount = len(ol_dict) + 1
    ol = _silo.DBOptlist(optcount, optcount * 150)

    for key, value in ol_dict.items():
        if isinstance(value, int):
            ol.add_int_option(key, value)
        elif isinstance(value, tuple):
            for el in value:
                if not isinstance(el, int):
                    raise TypeError("For now only tuples of int are "
                            "implemented as option value!")
            ol.add_option(key, value)
        else:
            ol.add_option(key, value)

    return ol


class SiloFile(_silo.DBFile):
    """This class can be used in a Python 2.5 *with* statement."""
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __init__(self, pathname, create=True, mode=None,
            fileinfo="Created using PyVisfile",
            target=DB_LOCAL, filetype=None):
        if create:
            if mode is None:
                mode = DB_NOCLOBBER
            if filetype is None:
                filetype = DB_PDB
            _silo.DBFile.__init__(self, pathname, mode, target,
                    fileinfo, filetype)
        else:
            if mode is None:
                mode = DB_APPEND
            if filetype is None:
                filetype = DB_UNKNOWN
            _silo.DBFile.__init__(self, pathname, filetype, mode)

    def put_zonelist_2(self, names, nzones, ndims, nodelist, lo_offset, hi_offset,
            shapetype, shapesize, shapecounts, optlist=None):
        if optlist is None:
            optlist = {}

        _silo.DBFile.put_zonelist_2(self, names, nzones, ndims,
                nodelist, lo_offset, hi_offset,
                shapetype, shapesize, shapecounts, _convert_optlist(optlist))

    def put_ucdmesh(self, mname, coordnames, coords,
            nzones, zonel_name, facel_name,
            optlist=None):
        if optlist is None:
            optlist = {}

        _silo.DBFile.put_ucdmesh(self, mname, coordnames, coords,
            nzones, zonel_name, facel_name, _convert_optlist(optlist))

    def put_ucdvar1(self, vname, mname, vec, centering, optlist=None):
        if optlist is None:
            optlist = {}

        _silo.DBFile.put_ucdvar1(self, vname, mname, vec, centering,
                _convert_optlist(optlist))

    def put_ucdvar(self, vname, mname, varnames, vars,
            centering, optlist=None):
        if optlist is None:
            optlist = {}

        _silo.DBFile.put_ucdvar(self, vname, mname, varnames, vars, centering,
                _convert_optlist(optlist))

    def put_defvars(self, vname, vars):
        """Add an defined variable ("expression") to this database.

        The *vars* argument consists of a list of tuples of type
        ``(name, definition)`` or
        ``(name, definition, DB_VARTYPE_SCALAR | DB_VARTYPE_VECTOR)`` or even
        ``(name, definition, DB_VARTYPE_XXX, {options})``
        If the type is not specified, scalar is assumed.
        """

        _silo.DBFile.put_defvars(self, vname, vars)

    def put_pointmesh(self, mname, coords, optlist=None):
        if optlist is None:
            optlist = {}

        _silo.DBFile.put_pointmesh(self, mname, coords,
                _convert_optlist(optlist))

    def put_pointvar1(self, vname, mname, var, optlist=None):
        if optlist is None:
            optlist = {}

        _silo.DBFile.put_pointvar1(self, vname, mname, var,
                _convert_optlist(optlist))

    def put_pointvar(self, vname, mname, vars, optlist=None):
        if optlist is None:
            optlist = {}

        _silo.DBFile.put_pointvar(self, vname, mname, vars,
                _convert_optlist(optlist))

    def put_quadmesh(self, mname, coords, coordtype=DB_COLLINEAR, optlist=None):
        if optlist is None:
            optlist = {}

        _silo.DBFile.put_quadmesh(self, mname, coords, coordtype,
                _convert_optlist(optlist))

    def put_quadvar1(self, vname, mname, var, dims, centering, optlist=None):
        if optlist is None:
            optlist = {}

        _silo.DBFile.put_quadvar1(self, vname, mname, var, dims, centering,
                _convert_optlist(optlist))

    def put_quadvar(self, vname, mname, varnames, vars, dims, centering,
            optlist=None):
        if optlist is None:
            optlist = {}

        _silo.DBFile.put_quadvar(self, vname, mname,
                varnames, vars, dims, centering,
                _convert_optlist(optlist))

    def put_multimesh(self, mname, mnames_and_types, optlist=None):
        if optlist is None:
            optlist = {}

        _silo.DBFile.put_multimesh(self, mname,
                mnames_and_types, _convert_optlist(optlist))

    def put_multivar(self, vname, vnames_and_types, optlist=None):
        if optlist is None:
            optlist = {}

        _silo.DBFile.put_multivar(self, vname,
                vnames_and_types, _convert_optlist(optlist))

    def put_curve(self, curvename, xvals, yvals, optlist=None):
        if optlist is None:
            optlist = {}

        _silo.DBFile.put_curve(self, curvename, xvals, yvals,
                _convert_optlist(optlist))


class ParallelSiloFile:
    """A :class:`SiloFile` that automatically creates a parallel master file.

    This class is meant to be instantiated on every rank of an MPI
    computation. It creates one data file per rank, and it
    automatically chooses a rank that writes a master file.

    The contents of the master file is automatically built,
    without any further user intervention.

    A :file:`.silo` extension is automatically appended to *pathname*
    for the master file, as are rank numbers and the extension
    for each individual rank.

    This class can be used as a context manager in ``with`` statement.

    .. automethod:: __init__
    .. automethod:: close
    .. automethod:: put_zonelist
    .. automethod:: put_zonelist_2
    .. automethod:: put_ucdmesh
    .. automethod:: put_ucdvar1
    .. automethod:: put_ucdvar
    .. automethod:: put_defvars
    .. automethod:: put_pointmesh
    .. automethod:: put_pointvar1
    .. automethod:: put_pointvar
    .. automethod:: put_quadmesh
    .. automethod:: put_quadvar1
    .. automethod:: put_quadvar
    """

    def __init__(self, pathname, rank, ranks, *args, **kwargs):
        self.rank = rank
        self.ranks = ranks

        rank_pathname_pattern = "{}-{:05d}.silo"
        rank_pathname = rank_pathname_pattern.format(pathname, rank)

        self.data_file = SiloFile(rank_pathname, *args, **kwargs)

        if self.rank == self.ranks[0]:
            head_pathname = f"{pathname}.silo"
            self.master_file = SiloFile(head_pathname, *args, **kwargs)

            self.rank_filenames = [
                    rank_pathname_pattern.format(pathname, fn_rank)
                    for fn_rank in ranks]
        else:
            self.master_file = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self.data_file.close()
        if self.master_file:
            self.master_file.close()

    # -------------------------------------------------------------------------
    def put_zonelist(self, *args, **kwargs):
        self.data_file.put_zonelist(*args, **kwargs)

    def put_zonelist_2(self, *args, **kwargs):
        self.data_file.put_zonelist_2(*args, **kwargs)

    def put_ucdmesh(self, mname, coordnames, coords,
            nzones, zonel_name, facel_name, optlist):
        self.data_file.put_ucdmesh(mname, coordnames, coords,
            nzones, zonel_name, facel_name, optlist)

        self._added_mesh(mname, DBObjectType.DB_UCDMESH, optlist)

    def put_ucdvar1(self, vname, mname, vec, centering, optlist=None):
        if optlist is None:
            optlist = {}

        self.data_file.put_ucdvar1(vname, mname, vec, centering, optlist)
        self._added_variable(vname, DBObjectType.DB_UCDVAR, optlist)

    def put_ucdvar(self, vname, mname, varnames, vars,
            centering, optlist=None):
        if optlist is None:
            optlist = {}

        self.data_file.put_ucdvar(vname, mname, varnames, vars, centering)
        self._added_variable(vname, DBObjectType.DB_UCDVAR, optlist)

    def put_defvars(self, vname, vars):
        """Add an defined variable ("expression") to this database.

        The *vars* argument consists of a list of tuples of type
        ``(name, definition)``
        or
        ``(name, definition, DB_VARTYPE_SCALAR | DB_VARTYPE_VECTOR)``.
        or even
        ``(name, definition, DB_VARTYPE_XXX, {options})``.
        If the type is not specified, scalar is assumed.
        """
        if self.master_file is not None:
            self.master_file.put_defvars(vname, vars)

    def put_pointmesh(self, mname, coords, optlist=None):
        if optlist is None:
            optlist = {}

        self.data_file.put_pointmesh(mname, coords, optlist)
        self._added_mesh(mname, DBObjectType.DB_POINTMESH, optlist)

    def put_pointvar1(self, vname, mname, var, optlist=None):
        if optlist is None:
            optlist = {}

        self.data_file.put_pointvar1(vname, mname, var, optlist)
        self._added_variable(vname, DBObjectType.DB_POINTVAR, optlist)

    def put_pointvar(self, vname, mname, vars, optlist=None):
        if optlist is None:
            optlist = {}

        self.data_file.put_pointvar(vname, mname, vars, optlist)
        self._added_variable(vname, DBObjectType.DB_POINTVAR, optlist)

    def put_quadmesh(self, mname, coords, coordtype=DB_COLLINEAR, optlist=None):
        if optlist is None:
            optlist = {}

        self.data_file.put_quadmesh(mname, coords, coordtype, optlist)
        self._added_mesh(mname, DBObjectType.DB_QUADMESH, optlist)

    def put_quadvar1(self, vname, mname, var, dims, centering, optlist=None):
        if optlist is None:
            optlist = {}

        self.data_file.put_quadvar1(vname, mname, var, dims, centering, optlist)
        self._added_variable(vname, DBObjectType.DB_QUADVAR, optlist)

    def put_quadvar(self, vname, mname, varnames, vars, dims, centering,
            optlist=None):
        if optlist is None:
            optlist = {}

        self.data_file.put_quadvar(vname, mname, varnames, vars,
                dims, centering, optlist)
        self._added_variable(vname, DBObjectType.DB_QUADVAR, optlist)

    def _added_mesh(self, mname, type, optlist):
        if self.master_file:
            self.master_file.put_multimesh(mname,
                    [(f"{rank_fn}:{mname}", type)
                        for rank_fn in self.rank_filenames],
                    optlist)

    def _added_variable(self, vname, type, optlist):
        if self.master_file:
            self.master_file.put_multivar(vname,
                    [(f"{rank_fn}:{vname}", type)
                        for rank_fn in self.rank_filenames],
                    optlist)

# vim: foldmethod=marker



__copyright__ = "Copyright (C) 2020 Alexandru Fikl"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from pytools import (
        add_tuples,
        generate_nonnegative_integer_tuples_summing_to_at_most as gnitstam)


__doc__ = """
VTK High-Order Lagrange Elements
--------------------------------

The high-order elements are described in
`this blog post <https://blog.kitware.com/modeling-arbitrary-order-lagrange-finite-elements-in-the-visualization-toolkit/>`_.
The ordering of the element nodes is as follows:

    1. the vertices (in an order that matches the linear elements,
       e.g. :data:`~pyvisfile.vtk.VTK_TRIANGLE`).
    2. the interior edge (or face in 2D) nodes, i.e. without the endpoints
    3. the interior face (3D only) nodes, i.e. without the edge nodes.
    4. the remaining interior nodes.

For simplices, the interior nodes are defined recursively by using the same
rules. However, for box elements the interior nodes are just listed in
order, with the last coordinate moving slowest.

To a large extent, the VTK ordering matches the ordering used by ``gmsh`` and
described `here <https://gmsh.info/doc/texinfo/gmsh.html#Node-ordering>`_.

.. autofunction:: vtk_lagrange_simplex_node_tuples
.. autofunction:: vtk_lagrange_simplex_node_tuples_to_permutation

.. autofunction:: vtk_lagrange_quad_node_tuples
.. autofunction:: vtk_lagrange_quad_node_tuples_to_permutation
"""     # noqa


# {{{

def add_tuple_to_list(ary, x):
    return [add_tuples(x, y) for y in ary]

# }}}


# {{{ VTK_LAGRANGE_${SIMPLEX} (i.e. CURVE/TRIANGLE/TETRAHEDRON)

def vtk_lagrange_curve_node_tuples(order):
    return [(0,), (order,)] + [(i,) for i in range(1, order)]


def vtk_lagrange_triangle_node_tuples(order):
    nodes = []
    offset = (0, 0)

    if order < 0:
        return nodes

    while order >= 0:
        if order == 0:
            nodes += [offset]
            break

        #   y
        #   ^
        #   |
        #   2
        #   |`\
        #   |  `\
        #   |    `\
        #   |      `\
        #   |        `\
        #   0----------1 --> x

        # add vertices
        vertices = [(0, 0), (order, 0), (0, order)]
        nodes += add_tuple_to_list(vertices, offset)
        if order == 1:
            break

        # add faces
        face_ids = range(1, order)
        faces = (
                # vertex 0 -> 1
                [(i, 0) for i in face_ids]
                # vertex 1 -> 2
                + [(order - i, i) for i in face_ids]
                # vertex 2 -> 0
                + [(0, order - i) for i in face_ids])
        nodes += add_tuple_to_list(faces, offset)

        order = order - 3
        offset = add_tuples(offset, (1, 1))

    return nodes


def vtk_lagrange_tetrahedron_node_tuples(order):
    nodes = []
    offset = (0, 0, 0)

    while order >= 0:
        if order == 0:
            nodes += [offset]
            break

        #                  z
        #               ,/
        #              3
        #            ,/|`\
        #          ,/  |  `\
        #        ,/    '.   `\
        #      ,/       |     `\
        #    ,/         |       `\
        #   0-----------'.--------2---> y
        #    `\.         |      ,/
        #       `\.      |    ,/
        #          `\.   '. ,/
        #             `\. |/
        #                `1
        #                  `.
        #                     x

        vertices = [(0, 0, 0), (order, 0, 0), (0, order, 0), (0, 0, order)]
        nodes += add_tuple_to_list(vertices, offset)
        if order == 1:
            break

        # add edges
        edge_ids = range(1, order)
        edges = (
                # vertex 0 -> 1
                [(i, 0, 0) for i in edge_ids]
                # vertex 1 -> 2
                + [(order - i, i, 0) for i in edge_ids]
                # vertex 2 -> 0
                + [(0, order - i, 0) for i in edge_ids]
                # vertex 0 -> 3
                + [(0, 0, i) for i in edge_ids]
                # vertex 1 -> 3
                + [(order - i, 0, i) for i in edge_ids]
                # vertex 2 -> 3
                + [(0, order - i, i) for i in edge_ids])
        nodes += add_tuple_to_list(edges, offset)

        # add faces
        face_ids = add_tuple_to_list(
                vtk_lagrange_triangle_node_tuples(order - 3), (1, 1))
        faces = (
                # face between vertices (0, 2, 3)
                [(i, 0, j) for i, j in face_ids]
                # face between vertices (1, 2, 3)
                + [(j, order - (i + j), i) for i, j in face_ids]
                # face between vertices (0, 1, 3)
                + [(0, j, i) for i, j in face_ids]
                # face between vertices (0, 1, 2)
                + [(j, i, 0) for i, j in face_ids])
        nodes += add_tuple_to_list(faces, offset)

        order = order - 4
        offset = add_tuples(offset, (1, 1, 1))

    return nodes


def vtk_lagrange_simplex_node_tuples(dims, order, vtk_version=(2, 1),
        is_consistent=None):
    """
    :arg dims: dimension of the simplex, i.e. 1 corresponds to a curve, 2 to
        a triangle, etc.
    :arg order: order of the polynomial representation, which also defines
        the number of nodes on the simplex.
    :arg vtk_version: a :class:`tuple` of two elements containing the version
        of the VTK XML file format in use. The ordering of some of the
        high-order elements changed between versions `2.1` and `2.2`.

    :return: a :class:`list` of ``dims``-dimensional tuples of integers
        up to ``order`` in the ordering expected by VTK. This list can be
        passed to :func:`vtk_lagrange_simplex_node_tuples_to_permutation`
        to obtain a permutation from the order used by :mod:`modepy`.
    """
    if dims == 1:
        return vtk_lagrange_curve_node_tuples(order)
    elif dims == 2:
        return vtk_lagrange_triangle_node_tuples(order)
    elif dims == 3:
        return vtk_lagrange_tetrahedron_node_tuples(order)
    else:
        raise ValueError(f"unsupported dimension: {dims}")


def vtk_lagrange_simplex_node_tuples_to_permutation(node_tuples):
    order = max([max(i) for i in node_tuples])
    dims = len(node_tuples[0])

    node_to_index = {
            node_tuple: i
            for i, node_tuple in enumerate(gnitstam(order, dims))
            }

    assert len(node_tuples) == len(node_to_index)
    return [node_to_index[v] for v in node_tuples]

# }}}


# {{{ VTK_LAGRANGE_${QUAD} (i.e. QUADRILATERAL/HEXAHEDRON)

def vtk_lagrange_quadrilateral_node_tuples(order):
    nodes = []

    if order < 0:
        return nodes

    if order == 0:
        return [(0, 0)]

    #   y
    #   ^
    #   |
    #   3----------2
    #   |          |
    #   |          |
    #   |          |
    #   |          |
    #   |          |
    #   0----------1 --> x

    # add vertices
    nodes += [(0, 0), (order, 0), (order, order), (0, order)]
    if order == 1:
        return nodes

    # add faces
    face_ids = range(1, order)
    nodes += (
            # vertex 0 -> 1
            [(i, 0) for i in face_ids]
            # vertex 1 -> 2
            + [(order, i) for i in face_ids]
            # vertex 2 -> 3
            + [(i, order) for i in face_ids]
            # vertex 3 -> 0
            + [(0, i) for i in face_ids])

    # add remaining interior nodes
    from itertools import product
    nodes += [(i, j) for j, i in product(range(1, order), repeat=2)]

    return nodes


def vtk_lagrange_hexahedon_node_tuples(order, vtk_version=(2, 1)):
    nodes = []

    if order < 0:
        return nodes

    if order == 0:
        return [(0, 0, 0)]

    #   z
    #   ^
    #   |
    #   4----------7
    #   |\         |\
    #   | \        | \
    #   |  \       |  \
    #   |   5------+---6
    #   |   |      |   |
    #   0---+------3---|--> y
    #    \  |       \  |
    #     \ |        \ |
    #      \|         \|
    #       1----------2
    #        \
    #         v x

    # add vertices
    nodes += [
            # (0, 1, 2, 3)
            (0, 0, 0), (order, 0, 0),
            (order, order, 0), (0, order, 0),
            # (4, 5, 6, 7)
            (0, 0, order), (order, 0, order),
            (order, order, order), (0, order, order),
            ]
    if order == 1:
        return nodes

    # add edges
    edge_ids = range(1, order)
    nodes += (
            # vertex 0 -> 1
            [(i, 0, 0) for i in edge_ids]
            # vertex 1 -> 2
            + [(order, i, 0) for i in edge_ids]
            # vertex 2 -> 3
            + [(i, order, 0) for i in edge_ids]
            # vertex 3 -> 0
            + [(0, i, 0) for i in edge_ids]

            # vertex 4 -> 5
            + [(i, 0, order) for i in edge_ids]
            # vertex 5 -> 6
            + [(order, i, order) for i in edge_ids]
            # vertex 6 -> 7
            + [(i, order, order) for i in edge_ids]
            # vertex 7 -> 4
            + [(0, i, order) for i in edge_ids]

            # vertex 0 -> 4
            + [(0, 0, i) for i in edge_ids]
            # vertex 1 -> 5
            + [(order, 0, i) for i in edge_ids]
            )

    if vtk_version <= (2, 1):
        nodes += (
            # vertex 3 -> 7
            [(0, order, i) for i in edge_ids]
            # vertex 2 -> 6
            + [(order, order, i) for i in edge_ids]
            )
    else:
        nodes += (
            # vertex 2 -> 6
            [(order, order, i) for i in edge_ids]
            # vertex 3 -> 7
            + [(0, order, i) for i in edge_ids]
            )

    # add faces
    from itertools import product
    nodes += (
            # face between (0, 4, 7, 3)
            [(0, i, j) for j, i in product(range(1, order), repeat=2)]
            # face between (1, 5, 6, 2)
            + [(order, i, j) for j, i in product(range(1, order), repeat=2)]
            # face between (0, 1, 5, 4)
            + [(i, 0, j) for j, i in product(range(1, order), repeat=2)]
            # face between (3, 2, 6, 7)
            + [(i, order, j) for j, i in product(range(1, order), repeat=2)]
            # face between (0, 1, 2, 3)
            + [(i, j, 0) for j, i in product(range(1, order), repeat=2)]
            # face between (4, 5, 6, 7)
            + [(i, j, order) for j, i in product(range(1, order), repeat=2)]
            )

    # add interior
    nodes += [(i, j, k) for k, j, i in product(range(1, order), repeat=3)]

    return nodes


def vtk_lagrange_quad_node_tuples(dims, order, vtk_version=(2, 1)):
    """
    :arg dims: dimension of the box, i.e. 1 corresponds to a curve, 2 to
        a quadrilateral, etc.
    :arg order: order of the polynomial representation, which also defines
        the number of nodes on the box.
    :arg vtk_version: a :class:`tuple` of two elements containing the version
        of the VTK XML file format in use. The ordering of some of the
        high-order elements changed between versions `2.1` and `2.2`.

    :return: a :class:`list` of ``dims``-dimensional tuples of integers
        up to ``order`` in the ordering expected by VTK. This list can be
        passed to :func:`vtk_lagrange_quad_node_tuples_to_permutation`
        to obtain a permutation from the order used by :mod:`modepy`.
    """
    if dims == 1:
        return vtk_lagrange_curve_node_tuples(order)
    elif dims == 2:
        return vtk_lagrange_quadrilateral_node_tuples(order)
    elif dims == 3:
        return vtk_lagrange_hexahedon_node_tuples(order, vtk_version=vtk_version)
    else:
        raise ValueError(f"unsupported dimension: {dims}")


def vtk_lagrange_quad_node_tuples_to_permutation(node_tuples):
    order = max([max(i) for i in node_tuples])
    dims = len(node_tuples[0])

    from itertools import product
    node_to_index = {
            node_tuple: i
            for i, node_tuple in enumerate(product(range(order + 1), repeat=dims))
            }

    assert len(node_tuples) == len(node_to_index)
    return [node_to_index[v] for v in node_tuples]

# }}}



import numpy as np
from pyvisfile.vtk import (
    UnstructuredGrid, DataArray,
    AppendedDataXMLGenerator,
    VTK_VERTEX, VF_LIST_OF_VECTORS, VF_LIST_OF_COMPONENTS)

n = 5000
points = np.random.randn(n, 3)

data = [
        ("p", np.random.randn(n)),
        ("vel", np.random.randn(3, n)),
]
file_name = "points.vtu"
compressor = None

grid = UnstructuredGrid(
        (n, DataArray("points", points, vector_format=VF_LIST_OF_VECTORS)),
        cells=np.arange(n, dtype=np.uint32),
        cell_types=np.asarray([VTK_VERTEX] * n, dtype=np.uint8))

for name, field in data:
    grid.add_pointdata(DataArray(name, field,
        vector_format=VF_LIST_OF_COMPONENTS))

from os.path import exists
if exists(file_name):
    raise RuntimeError("output file '%s' already exists"
        % file_name)

outf = open(file_name, "w")
AppendedDataXMLGenerator(compressor)(grid).write(outf)
outf.close()



__copyright__ = "Copyright (C) 2020 Alexandru Fikl"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import os
import enum

from typing import Any, Dict, Optional, Tuple, Union
from xml.etree.ElementTree import Element, ElementTree

import numpy as np

__doc__ = """
Xdmf Tags
---------

DataItem
^^^^^^^^

.. autoclass:: DataItemType
    :show-inheritance:
    :members:
    :undoc-members:

.. autoclass:: DataItemNumberType
    :show-inheritance:
    :members:
    :undoc-members:
.. autoclass:: DataItemFormat
    :show-inheritance:
    :members:
    :undoc-members:
.. autoclass:: DataItemEndian
    :show-inheritance:
    :members:
    :undoc-members:

.. autoclass:: DataItem

Domain
^^^^^^

.. autoclass:: Domain

Grid
^^^^

.. autoclass:: GridType
    :show-inheritance:
    :members:
    :undoc-members:
.. autoclass:: CollectionType
    :show-inheritance:
    :members:
    :undoc-members:
.. autoclass:: Grid


Topology
^^^^^^^^

.. autoclass:: TopologyType
    :show-inheritance:
    :members:
    :undoc-members:
.. autoclass:: Topology

Geometry
^^^^^^^^

.. autoclass:: GeometryType
    :show-inheritance:
    :members:
    :undoc-members:
.. autoclass:: Geometry

Attribute
^^^^^^^^^

.. autoclass:: AttributeType
    :show-inheritance:
    :members:
    :undoc-members:
.. autoclass:: AttributeCenter
    :show-inheritance:
    :members:
    :undoc-members:
.. autoclass:: Attribute

Time
^^^^

.. autoclass:: Time
    :members:

Information
^^^^^^^^^^^

.. autoclass:: Information
    :members:

XInclude
^^^^^^^^

.. autoclass:: XInclude
    :members:

Writing
-------

.. autoclass:: DataArray
.. autoclass:: NumpyDataArray

.. autoclass:: XdmfGrid
.. autoclass:: XdmfUnstructuredGrid
    :show-inheritance:

.. autoclass:: XdmfWriter
"""


# {{{ utils

def _stringify(obj):
    if isinstance(obj, (list, tuple)):
        return " ".join(str(c) for c in obj)
    return str(obj)


class XdmfElement(Element):
    """Base class for all the XDMF tags.

    .. attribute:: name

        The ``Name`` attribute the tags. This attribute is optional, so
        it can return *None* if not set.

    .. automethod:: replace
    """

    def __init__(self,
            parent: Optional[Element],
            tag: str,
            attrib: Dict[str, Optional[Any]]):
        super().__init__(tag, attrib={
            k: _stringify(v) for k, v in attrib.items() if v is not None
            })

        self.parent = parent
        if parent is not None:
            parent.append(self)

    @property
    def name(self):
        return self.get("Name")

    def replace(self, **kwargs):
        """Duplicate the current tag with updated attributes from *kwargs*."""
        parent = kwargs.pop("parent", self.parent)
        tag = kwargs.pop("tag", self.tag)

        attrib = self.attrib.copy()
        attrib.update(kwargs)

        return XdmfElement.__init__(self, parent, tag, attrib=attrib)

# }}}


# {{{ xdmf tags

# {{{ attribute

# {{{ unsupported

# NOTE: these are taken from VTK source in
#
#   https://gitlab.kitware.com/vtk/vtk/-/blob/c138bfae93f570b467e7f4fdc0c42d974cd684f4/IO/Xdmf3/vtkXdmf3DataSet.cxx#L2278
#
# and are currently unsupported here (someone needs to figure out how to write
# them). They were introduced in
#
#   https://gitlab.kitware.com/xdmf/xdmf/-/merge_requests/41
#   https://gitlab.kitware.com/vtk/vtk/-/merge_requests/3194

class AttributeElementFamily(enum.Enum):
    """High-order element families for ``FiniteElementFunction``."""
    DG = enum.auto()    #: Discontinuous Galerkin elements for simplices.
    CG = enum.auto()    #: Continuous Galerkin elements for simplices.
    Q = enum.auto()     #: Continuous Galerkin elements for quadrilaterals.
    DQ = enum.auto()    #: Discontinuous Galerkin elements for quadrilaterals.
    RT = enum.auto()    #: Raviart-Thomas elements on triangles.


class AttributeElementCell(enum.Enum):
    """Reference element types for ``FiniteElementFunction``."""
    interval = enum.auto()
    triangle = enum.auto()
    tetrahedron = enum.auto()
    quadrilateral = enum.auto()
    hexahedron = enum.auto()

# }}}


class AttributeType(enum.Enum):
    """Rank of the attribute stored on the mesh."""
    # NOTE: integer ids taken from
    # https://gitlab.kitware.com/xdmf/xdmf/-/blob/04a84bab0eb1568e0f1a27c8fb60c6931efda003/XdmfAttributeType.hpp#L129
    Scalar = 200
    Vector = 201
    Tensor = 202
    Matrix = 203
    Tensor6 = 204
    GlobalId = 205

    @staticmethod
    def from_shape(shape: Tuple[int, ...]) -> "AttributeType":
        # https://github.com/nschloe/meshio/blob/37673c8fb938ad73d92fb3171dee3eb193b5e7ac/meshio/xdmf/common.py#L162
        if len(shape) == 1 or (len(shape) == 2 and shape[1] == 1):
            return AttributeType.Scalar
        elif len(shape) == 2 and shape[0] in [2, 3]:
            return AttributeType.Vector
        elif len(shape) == 2 and shape[0] in [4, 9]:
            return AttributeType.Tensor
        elif len(shape) == 2 and shape[0] == 6:
            return AttributeType.Tensor6
        elif len(shape) == 3:
            return AttributeType.Matrix
        else:
            raise ValueError(f"cannot determine attribute type from shape '{shape}'")


class AttributeCenter(enum.Enum):
    """Center of the attribute stored on the mesh."""
    # NOTE: integer ids taken from
    # https://gitlab.kitware.com/xdmf/xdmf/-/blob/04a84bab0eb1568e0f1a27c8fb60c6931efda003/XdmfAttributeCenter.hpp#L126
    Grid = 100
    Cell = 101
    Face = 102
    Edge = 103
    Node = 104
    # NOTE: only for `FiniteElementFunction` attributes, which are not supported
    Other = 105


class Attribute(XdmfElement):
    """
    .. automethod:: __init__
    """

    def __init__(
            self, *,
            name: Optional[str] = None,
            atype: AttributeType = AttributeType.Scalar,
            acenter: AttributeCenter = AttributeCenter.Node,
            parent: Optional[Element] = None,
            ):
        """
        :param parent: if provided, *self* is appended to the element.
        """

        super().__init__(parent, "Attribute", {
            "Name": name,
            "Center": acenter.name,
            "AttributeType": atype.name,
            })

# }}}


# {{{ data item

class DataItemType(enum.Enum):
    """Data layout of an item."""
    Uniform = enum.auto()
    HyperSlab = enum.auto()
    Function = enum.auto()


class DataItemNumberType(enum.Enum):
    """Basic number types for an item."""
    Char = enum.auto()
    UChar = enum.auto()
    Int = enum.auto()
    UInt = enum.auto()
    Float = enum.auto()

    @staticmethod
    def from_dtype(dtype: np.dtype) -> "DataItemNumberType":
        if dtype.kind == "i":
            return DataItemNumberType.Int
        elif dtype.kind == "u":
            return DataItemNumberType.UInt
        elif dtype.kind == "f":
            return DataItemNumberType.Float
        else:
            raise ValueError(f"unsupported dtype: '{dtype}'")


class DataItemFormat(enum.Enum):
    """Format in which the item is stored."""
    XML = enum.auto()
    HDF = enum.auto()
    Binary = enum.auto()
    # NOTE: unsupported as it's pretty much undocumented
    TIFF = enum.auto()


class DataItemEndian(enum.Enum):
    """Endianess of the data stored in the item."""
    # NOTE: integer ids taken from
    # https://gitlab.kitware.com/xdmf/xdmf/-/blob/04a84bab0eb1568e0f1a27c8fb60c6931efda003/core/XdmfBinaryController.hpp#L211
    Big = 50
    Little = 51
    Native = 52

    @staticmethod
    def from_system() -> "DataItemEndian":
        import sys
        if sys.byteorder == "little":
            return DataItemEndian.Little
        elif sys.byteorder == "big":
            return DataItemEndian.Big
        else:
            return DataItemEndian.Native


class DataItem(XdmfElement):
    """A :class:`DataItem` describes the storage of actual values in an
    XDMF file. This can be inline ASCII data, the path to a binary file
    or a reference to another :class:`DataItem`.

    .. attribute:: dimensions

        Analogous to :attr:`numpy.ndarray.shape`.

    .. automethod:: __init__
    .. automethod:: as_reference
    """

    def __init__(
            self, *,
            dimensions: Optional[Tuple[int, ...]] = None,
            name: Optional[str] = None,
            itype: Optional[DataItemType] = DataItemType.Uniform,
            ntype: Optional[DataItemNumberType] = DataItemNumberType.Float,
            precision: Optional[int] = 4,
            reference: Optional[str] = None,
            function: Optional[str] = None,
            endian: Optional[DataItemEndian] = DataItemEndian.Native,
            dformat: Optional[DataItemFormat] = DataItemFormat.XML,
            parent: Optional[Element] = None,
            data: Optional[str] = None,
            ):
        """
        :param parent: if provided, *self* is appended to the element.
        :param reference: path to another :class:`DataItem`.
            Use :meth:`as_reference` to populate.
        :param data: data contained inside the :class:`DataItem`. This is
            usually a path to a binary file.
        """

        self._dimensions = dimensions

        super().__init__(parent, "DataItem", {
            "Name": name,
            "ItemType": itype.name if itype is not None else itype,
            "Dimensions": dimensions,
            "NumberType": ntype.name if ntype is not None else ntype,
            "Precision": precision,
            "Reference": reference,
            "Function": function,
            "Endian": endian.name if endian is not None else endian,
            "Format": dformat.name if dformat is not None else dformat,
            })

        if data is not None:
            self.text = data

    @property
    def dimensions(self):
        if self._dimensions is None:
            raise AttributeError

        return self._dimensions

    @classmethod
    def as_reference(cls, reference_name: str, *,
            parent: Optional[Element] = None) -> "DataItem":
        """
        :param reference_name: a name or an absolute reference to another
            :class:`DataItem`. The name is just the ``Name`` attribute of
            the item, which is assumed to be in the top :class:`Domain`. If
            another :class:`DataItem` needs to be references, or there are
            multiple domains, use an absolute reference path, as defined
            in the `XDMF docs <https://www.xdmf.org/index.php/XDMF_Model_and_Format>`__.
        """     # noqa: E501

        if not reference_name.startswith("/"):
            reference = f"/Xdmf/Domain/DataItem[@Name='{reference_name}']"
        else:
            reference = reference_name

        return cls(
                reference="XML",
                data=reference,
                itype=None, ntype=None, precision=None,
                endian=None, dformat=None,
                parent=parent)


def _data_item_format_from_str(text: str) -> DataItemFormat:
    # NOTE: this handles two cases (not meant to be too smart about it)
    # 1. if the text is just `some_file.bin` or `some_file.out`, we assume it's
    #    a custom binary file.
    # 2. if the text is `some_file.h5:/group/dataset`, we assume it's an HDF5
    #    file.
    if ":" not in text:
        return DataItemFormat.Binary

    try:
        filename, _ = text.split(":")
    except ValueError as exc:
        raise ValueError("cannot determine format from text") from exc

    if filename.endswith(".h5"):
        return DataItemFormat.HDF
    else:
        raise ValueError("cannot determine format from text")


def _data_item_from_numpy(
        ary: np.ndarray, *,
        name: Optional[str] = None,
        parent: Optional[Element] = None,
        data: Optional[str] = None,
        dformat: Optional[DataItemFormat] = None) -> DataItem:
    """Create a :class:`DataItem` from a given :class:`~numpy.ndarray`.

    .. note::

        This is meant for internal use only. Use :class:`NumpyDataArray` instead.
    """

    if dformat is None:
        if data is None:
            dformat = DataItemFormat.XML
        else:
            dformat = _data_item_format_from_str(data)

    return DataItem(
            name=name,
            dimensions=ary.shape,
            itype=DataItemType.Uniform,
            ntype=DataItemNumberType.from_dtype(ary.dtype),
            precision=ary.dtype.itemsize,
            endian=DataItemEndian.from_system(),
            dformat=dformat,
            parent=parent,
            data=data,
            )


def _join_data_items(
        items: Tuple[DataItem, ...], *,
        parent: Optional[Element] = None) -> DataItem:
    r"""Joins several :class:`DataItem`\ s using a :attr:`DataItemType.Function`
    as::

        JOIN($0, $1, ...)

    (Used for describing vectors from scalar data.)
    See the `Xdmf Function docs <https://www.xdmf.org/index.php/XDMF_Model_and_Format#Function>`__
    for more information.

    :returns: the newly created :class:`DataItem` that joins the input items.
    """     # noqa: E501

    if len(items) == 1:
        item = items[0]
    else:
        from pytools import is_single_valued
        if not is_single_valued(item.dimensions for item in items):
            raise ValueError("items must have the same dimension")

        dimensions = (len(items),) + items[0].dimensions
        ids = ", ".join(f"${i}" for i in range(dimensions[0]))

        item = DataItem(
                dimensions=dimensions,
                itype=DataItemType.Function,
                ntype=None,
                precision=None,
                function=f"JOIN({ids})",
                endian=None,
                dformat=None,
                )

        for subitem in items:
            item.append(subitem)

    if parent is not None:
        parent.append(item)

    return item

# }}}


# {{{ grid

class GridType(enum.Enum):
    """General structure of the connectivity."""
    Uniform = enum.auto()
    Collection = enum.auto()
    Tree = enum.auto()
    SubSet = enum.auto()


class CollectionType(enum.Enum):
    Spatial = enum.auto()
    Temporal = enum.auto()


class Grid(XdmfElement):
    """
    .. automethod:: __init__
    """

    def __init__(
            self, *,
            name: Optional[str] = None,
            gtype: GridType = GridType.Uniform,
            ctype: Optional[CollectionType] = None,
            parent: Optional[Element] = None,
            ):
        """
        :param parent: if provided, *self* is appended to the element.
        """

        if gtype == GridType.Collection and ctype is None:
            ctype = CollectionType.Spatial

        super().__init__(parent, "Grid", {
            "Name": name,
            "GridType": gtype.name,
            "CollectionType": ctype.name if ctype is not None else ctype,
            })

# }}}


# {{{ topology

class TopologyType(enum.IntEnum):
    """Element and mesh layouts."""
    NoTopology = 0x0
    Mixed = 0x70

    # linear elements from XdmfTopologyType.cpp
    Polyvertex = 0x1
    Polyline = 0x2
    Polygon = 0x3
    Triangle = 0x4
    Quadrilateral = 0x5
    Tetrahedron = 0x6
    Pyramid = 0x7
    Wedge = 0x8
    Hexahedron = 0x9
    Polyhedron = 0x10
    # quadratic elements from XdmfTopologyType.cpp
    Edge_3 = 0x22
    Triangle_6 = 0x24
    Quadrilateral_8 = 0x25
    Tetrahedron_10 = 0x26
    Pyramid_13 = 0x27
    Wedge_15 = 0x28
    Wedge_18 = 0x29
    Hexahedron_20 = 0x30
    # high-order elements from XdmfTopologyType.cpp
    Hexahedron_24 = 0x31
    Hexahedron_27 = 0x32
    Hexahedron_64 = 0x33
    Hexahedron_125 = 0x34
    Hexahedron_216 = 0x35
    Hexahedron_343 = 0x36
    Hexahedron_512 = 0x37
    Hexahedron_729 = 0x38
    Hexahedron_1000 = 0x39
    Hexahedron_1331 = 0x40
    # spectral elements from XdmfTopologyType.cpp
    Hexahedron_Spectral_64 = 0x41
    Hexahedron_Spectral_125 = 0x42
    Hexahedron_Spectral_216 = 0x43
    Hexahedron_Spectral_343 = 0x44
    Hexahedron_Spectral_512 = 0x45
    Hexahedron_Spectral_729 = 0x46
    Hexahedron_Spectral_1000 = 0x47
    Hexahedron_Spectral_1331 = 0x48

    # structured from XdmfCurvilinearGrid.cpp
    SMesh2D = 0x1110        # Curvilinear mesh
    SMesh3D = 0x1110
    # structured from XdfmRectilinearGrid.cpp
    RectMesh2D = 0x1101     # Mesh with perpendicular axes
    RectMesh3D = 0x1101
    # stuctured from XdmfRegularGrid.cpp
    CoRectMesh2D = 0x1102   # Mesh with equally spaced perpendicular axes
    CoRectMesh3D = 0x1102


class Topology(XdmfElement):
    """
    .. automethod:: __init__
    """

    def __init__(
            self, *,
            ttype: TopologyType,
            nodes_per_element: Optional[int] = None,
            number_of_elements: Optional[int] = None,
            dimensions: Optional[Tuple[int, ...]] = None,
            parent: Optional[Element] = None,
            ):
        """
        :param parent: if provided, *self* is appended to the element.
        """

        ttype_name = _XDMF_TOPOLOGY_TYPE_TO_NAME.get(ttype, ttype.name)
        if ttype in _XDMF_ELEMENT_NODE_COUNT or ttype in _XDMF_STRUCTURED_GRIDS:
            if nodes_per_element is not None:
                raise ValueError(f"cannot set 'nodes_per_element' for {ttype_name}")
        else:
            if nodes_per_element is None:
                raise ValueError(f"'nodes_per_element' required for {ttype_name}")

        super().__init__(parent, "Topology", {
            "TopologyType": ttype_name,
            "Dimensions": dimensions,
            "NodesPerElement": nodes_per_element,
            "NumberOfElements": number_of_elements,
            })


_XDMF_ELEMENT_NODE_COUNT = {
        TopologyType.Polyvertex: 1,
        # TopologyType.Polyline: user-defined
        # TopologyType.Polygon: user-defined
        TopologyType.Triangle: 3,
        TopologyType.Quadrilateral: 4,
        TopologyType.Tetrahedron: 4,
        TopologyType.Pyramid: 5,
        TopologyType.Wedge: 6,
        TopologyType.Hexahedron: 8,
        # TopologyType.Polyhedron: user-defined
        TopologyType.Edge_3: 3,
        TopologyType.Triangle_6: 6,
        TopologyType.Quadrilateral_8: 8,
        TopologyType.Tetrahedron_10: 10,
        TopologyType.Pyramid_13: 13,
        TopologyType.Wedge_15: 15,
        TopologyType.Wedge_18: 18,
        TopologyType.Hexahedron_20: 20,
        TopologyType.Hexahedron_24: 24,
        TopologyType.Hexahedron_27: 27,
        TopologyType.Hexahedron_64: 64,
        TopologyType.Hexahedron_125: 125,
        TopologyType.Hexahedron_216: 216,
        TopologyType.Hexahedron_343: 343,
        TopologyType.Hexahedron_512: 512,
        TopologyType.Hexahedron_729: 729,
        TopologyType.Hexahedron_1000: 1000,
        TopologyType.Hexahedron_1331: 1331,
        TopologyType.Hexahedron_Spectral_64: 64,
        TopologyType.Hexahedron_Spectral_125: 125,
        TopologyType.Hexahedron_Spectral_216: 216,
        TopologyType.Hexahedron_Spectral_343: 343,
        TopologyType.Hexahedron_Spectral_512: 512,
        TopologyType.Hexahedron_Spectral_729: 729,
        TopologyType.Hexahedron_Spectral_1000: 1000,
        TopologyType.Hexahedron_Spectral_1331: 1331,
        }


_XDMF_STRUCTURED_GRIDS = {
        TopologyType.SMesh2D,
        TopologyType.SMesh3D,
        TopologyType.RectMesh2D,
        TopologyType.RectMesh3D,
        TopologyType.CoRectMesh2D,
        TopologyType.CoRectMesh3D,
        }


# NOTE: the names in TopologyType are weird because python identifiers cannot
# start with a number, so 2DSMesh is not allowed
_XDMF_TOPOLOGY_TYPE_TO_NAME = {
        TopologyType.SMesh2D: "2DSMesh",
        TopologyType.SMesh3D: "3DSMesh",
        TopologyType.RectMesh2D: "2DRectMesh",
        TopologyType.RectMesh3D: "3DRectMesh",
        TopologyType.CoRectMesh2D: "2DCoRectMesh",
        TopologyType.CoRectMesh3D: "3DCoRectMesh",
        }

# }}}


# {{{ geometry


class GeometryType(enum.Enum):
    """Data layout of the node coordinates."""
    XY = enum.auto()
    XYZ = enum.auto()
    # NOTE: all of these don't seem to be supported in VTK/Paraview with
    # XDMF3 for some reason, but are still mentioned in the XDMF source
    VXVY = enum.auto()
    VXVYVZ = enum.auto()
    ORIGIN_DXDY = enum.auto()
    ORIGIN_DXDYDZ = enum.auto()


class Geometry(XdmfElement):
    """
    .. automethod:: __init__
    """

    def __init__(
            self, *,
            name: Optional[str] = None,
            gtype: GeometryType = GeometryType.XYZ,
            parent: Optional[Element] = None,
            ):
        """
        :param parent: if provided, *self* is appended to the element.
        """

        super().__init__(parent, "Geometry", {
            "Name": name,
            "GeometryType": gtype.name,
            })

# }}}


# {{{ time

class Time(XdmfElement):
    """
    .. automethod:: __init__
    """

    def __init__(
            self, *,
            value: str,
            parent: Optional[Element] = None,
            ):
        """
        :param parent: if provided, *self* is appended to the element.
        """

        super().__init__(parent, "Time", {
            "Value": value,
            })

# }}}


# {{{ domain

class Domain(XdmfElement):
    """
    .. automethod:: __init__
    """

    def __init__(
            self, *,
            name: Optional[str] = None,
            parent: Optional[Element] = None,
            ):
        """
        :param parent: if provided, *self* is appended to the element.
        """

        super().__init__(parent, "Domain", {
            "Name": name,
            })

# }}}


# {{{ information

class Information(XdmfElement):
    """
    .. automethod:: __init__
    """

    def __init__(
            self, *,
            name: str,
            value: str,
            parent: Optional[Element] = None,
            ):
        """
        :param parent: if provided, *self* is appended to the element.
        """

        super().__init__(parent, "Information", {
            "Name": name,
            "Value": value,
            })

# }}}


# {{{ include

class XInclude(XdmfElement):
    """
    .. automethod:: __init__
    """

    def __init__(
            self, *,
            href: Optional[str],
            xpointer: Optional[str] = None,
            parent: Optional[Element] = None,
            ):
        """
        :param parent: if provided, *self* is appended to the element.
        :param xpointer: path inside the file represented by *href*.
        """
        if xpointer is not None:
            xpointer = f"xpointer({xpointer})"

        super().__init__(parent, "xi:include", {
            "href": href,
            "xpointer": xpointer,
            })

# }}}

# }}}


# {{{ data arrays

def _ndarray_to_string(ary):
    if not isinstance(ary, np.ndarray):
        raise TypeError(f"expected an 'ndarray', got '{type(ary).__name__}'")

    ntype = DataItemNumberType.from_dtype(ary.dtype)
    if ntype == DataItemNumberType.Int or ntype == DataItemNumberType.UInt:
        fmt = "%d"
    elif ntype == DataItemNumberType.Float:
        if ary.dtype.itemsize == 8:
            fmt = "%.16e"
        elif ary.dtype.itemsize == 4:
            fmt = "%.8e"
        else:
            raise ValueError(f"unsupported dtype item size: {ary.dtype.itemsize}")
    else:
        raise ValueError(f"unsupported dtype: '{ary.dtype}'")

    import io
    bio = io.BytesIO()
    np.savetxt(bio, ary, fmt=fmt)

    return "\n" + bio.getvalue().decode()


def _geometry_type_from_points(points):
    dims = points.shape[-1]

    if len(points.components) == 1:
        if dims == 2:
            return GeometryType.XY
        elif dims == 3:
            return GeometryType.XYZ
        else:
            raise ValueError(f"unsupported dimension: '{dims}'")
    else:
        if dims == 2:
            return GeometryType.VXVY
        elif dims == 3:
            return GeometryType.VXVYVZ
        else:
            raise ValueError(f"unsupported dimension: '{dims}'")


class DataArray:
    r"""An array represented as a list of :class:`DataItem`\ s.

    .. automethod:: __init__
    .. automethod:: as_data_item
    """

    def __init__(
            self,
            components: Tuple[DataItem, ...], *,
            name: Optional[str] = None,
            acenter: Optional[AttributeCenter] = None,
            atype: Optional[AttributeType] = None,
            ):
        r"""
        :param components: a description of each component of an array.
        :param name: name of the array. This name will be used if the array
            is added as an attribute, otherwise the names of the *components*
            are used.
        """
        if not isinstance(components, tuple):
            raise TypeError("'components' should be a tuple")

        if name is None:
            if len(components) == 1:
                name = components[0].name
            else:
                # NOTE: assumes component names are {name}_{suffix}
                name = os.path.commonprefix([
                    c.name for c in components
                    ]).strip("_")

        self.components = components

        self.name = name
        self.acenter = acenter
        self.atype = atype

    @property
    def shape(self):
        if len(self.components) == 1:
            return self.components[0].dimensions
        else:
            return (len(self.components),) + self.components[0].dimensions

    def as_data_item(self, *,
            parent: Optional[Element] = None) -> Tuple[DataItem, ...]:
        r"""Finalize the :class:`DataArray` and construct :class:`DataItem`\ s
        to be written to a file.
        """

        items = self.components[:]
        if parent is not None:
            for item in items:
                if parent.tag != "Attribute":
                    item.set("Name", self.name)

                parent.append(item)

        return items

    @classmethod
    def from_dataset(cls, dset,
            acenter: AttributeCenter = AttributeCenter.Node,
            atype: Optional[AttributeType] = None) -> "DataArray":
        filename = dset.file.filename
        data = f"{filename}:{dset.name}"
        name = dset.name.split("/")[-1]

        item = _data_item_from_numpy(dset,
                data=data,
                dformat=DataItemFormat.HDF)

        return cls(
                components=(item,),
                name=name,
                acenter=acenter,
                atype=atype)


class NumpyDataArray(DataArray):
    """
    .. automethod:: __init__
    """

    def __init__(
            self,
            ary: np.ndarray, *,
            acenter: Optional[AttributeCenter] = None,
            name: Optional[str] = None,
            ):
        """
        :param ary: if this is an :class:`object` array, each entry is considered
            a different component and will consist of a separate
            :class:`DataItem`.
        """

        if ary.dtype.char == "O":
            from pytools import is_single_valued
            if not is_single_valued(iary.shape for iary in ary):
                raise ValueError("'ary' components must have the same size")

            items = tuple([
                    _data_item_from_numpy(iary, name=f"{name}_{i}")
                    for i, iary in enumerate(ary)
                    ])
        else:
            items = (_data_item_from_numpy(ary, name=name),)

        super().__init__(items, name=name, acenter=acenter)
        self.ary = ary

    def as_data_item(self, *,
            parent: Optional[Element] = None) -> Tuple[DataItem, ...]:
        items = super().as_data_item(parent=parent)

        if self.ary.dtype.char == "O":
            ary = tuple(self.ary)
        else:
            ary = (self.ary,)

        for item, iary in zip(items, ary):
            item.text = _ndarray_to_string(iary)

        return items

# }}}


# {{{ grids

class XdmfGrid:
    """
    .. automethod:: __init__
    .. automethod:: add_attribute
    """

    def __init__(self, root: Grid):
        self.root = root

    def getroot(self):
        return self.root

    def add_attribute(self, ary: DataArray, *, join: bool = True) -> Attribute:
        """
        :param ary:
        :param join: If *True* and *ary* has multiple components, they are
            joined using an XDMF Function.
        """

        acenter = ary.acenter
        if acenter is None:
            acenter = AttributeCenter.Node

        atype = ary.atype
        if atype is None:
            atype = AttributeType.from_shape(ary.shape[::-1])

        attr = Attribute(
                name=ary.name,
                atype=atype,
                acenter=acenter,
                parent=self.getroot(),
                )

        if join:
            items = ary.as_data_item()
            _join_data_items(items, parent=attr)
        else:
            ary.as_data_item(parent=attr)

        return attr


class XdmfUnstructuredGrid(XdmfGrid):
    """
    .. automethod:: __init__
    """

    def __init__(self,
            points: DataArray,
            connectivity: DataArray, *,
            topology_type: Union[Topology, TopologyType],
            name: Optional[str] = None,
            geometry_type: Optional[GeometryType] = None):
        if geometry_type is None:
            geometry_type = _geometry_type_from_points(points)

        if geometry_type not in (GeometryType.XY, GeometryType.XYZ):
            # NOTE: Paraview 5.8 seems confused when using VXVY geometry types
            raise ValueError(f"unsupported geometry type: '{geometry_type}'")

        grid = Grid(parent=None, name=name)

        nelements = np.prod(connectivity.shape[:-1])
        if isinstance(topology_type, TopologyType):
            if topology_type == TopologyType.Polyline:
                nodes_per_element = 2
            else:
                nodes_per_element = None

            topology = Topology(
                    parent=grid,
                    ttype=topology_type,
                    nodes_per_element=nodes_per_element,
                    number_of_elements=nelements)
        elif isinstance(topology, Topology):
            topology = topology_type.replace({
                "parent": grid,
                "number_of_elements": nelements
                })
        else:
            raise TypeError(f"unsupported type: {type(topology_type).__name__}")

        connectivity.as_data_item(parent=topology)

        geometry = Geometry(parent=grid, gtype=geometry_type)
        points.as_data_item(parent=geometry)

        super().__init__(grid)

# }}}


# {{{ writer

class XdmfWriter(ElementTree):
    """
    .. automethod:: __init__
    .. automethod:: write
    .. automethod:: write_pretty
    """

    def __init__(self,
            grids: Tuple[XdmfGrid, ...], *,
            arrays: Optional[Tuple[DataArray, ...]] = None,
            tags: Optional[Tuple[Element, ...]] = None):
        r"""
        :param grids: a :class:`tuple` of grids to be added to the
            top :class:`Domain`. Currently only a single domain is supported.
        :param arrays: additional :class:`DataArray`\ s to be added to the
            top :class:`Domain`, as opposed to as attribute on the grids.
        """
        root = Element("Xdmf", {
            "xmlns:xi": "http://www.w3.org/2001/XInclude",
            "Version": "3.0",
            })

        domain = Domain(parent=root)
        if arrays is not None:
            for ary in arrays:
                ary.as_data_item(parent=domain)

        if tags is not None:
            for tag in tags:
                domain.append(tag)

        for grid in grids:
            domain.append(grid.getroot())

        super().__init__(root)

    def write_pretty(self, filename):
        """Produces a nicer-looking XML file with clean indentation."""
        # https://stackoverflow.com/a/1206856
        from xml.etree.ElementTree import tostring
        from xml.dom import minidom
        dom = minidom.parseString(tostring(
            self.getroot(),
            encoding="utf-8",
            short_empty_elements=False,
            ))

        with open(filename, "wb") as fd:
            fd.write(dom.toprettyxml(indent="  ", encoding="utf-8"))

    def write(self, filename):
        """Write the the XDMF file."""
        super().write(
                filename,
                encoding="utf-8",
                xml_declaration=True,
                short_empty_elements=False,
                )

# }}}



from pyvisfile.vtk import write_structured_grid

import numpy as np

angle_mesh = np.mgrid[1:2:10j, 0:2*np.pi:20j]

r = angle_mesh[0, np.newaxis]
phi = angle_mesh[1, np.newaxis]
mesh = np.vstack((
    r*np.cos(phi),
    r*np.sin(phi),
    ))

from pytools.obj_array import make_obj_array
vec = make_obj_array([
    np.cos(phi),
    np.sin(phi),
    ])

write_structured_grid("yo-2d.vts", mesh,
        point_data=[("phi", phi), ("vec", vec)])



__author__ = "Christoph Statz, christoph.statz <at> tu-dresden.de"

import numpy as np

from pyvisfile.silo import SiloFile, \
        DB_HDF5, DB_CLOBBER, DBOPT_CYCLE, DBOPT_DTIME, DB_COLLINEAR, \
        DBOPT_XUNITS, DBOPT_YUNITS, DBOPT_XLABEL, DBOPT_YLABEL, \
        DBOPT_UNITS, DB_NODECENT, DBObjectType, DB_LOCAL, \
        DBOPT_HI_OFFSET


x_len = 50
y_len = 40
x = np.linspace(-10., 20., x_len)
y = np.linspace(20., 40., y_len)

# Funny, but in row-major y comes first.
data_global = np.zeros((y_len, x_len), dtype=np.float64, order="C")

x_from = [0, 19, 0, 19]
x_to = [21, 50, 21, 50]
y_from = [0, 0, 19, 19]
y_to = [21, 21, 40, 40]

# Mark the Ghostzones (at least in one direction).
hi_offset = [(1, 1), (0, 1), (1, 0), (0, 0)]

data_global[:20, :20] = 0.
data_global[20:, :20] = 1.
data_global[:20, 20:] = 2.
data_global[20:, 20:] = 3.

mesh_names = []
var_names = []

for i in range(4):

    file_name = "example_%05d.silo" % i
    s = SiloFile(file_name,
            mode=DB_CLOBBER,
            filetype=DB_HDF5,
            target=DB_LOCAL,
            fileinfo=f"Example Silo {i:05d}.")

    axes = (x[x_from[i]:x_to[i]], y[y_from[i]:y_to[i]])

    # Copy necessary due to slicing!
    data = data_global[y_from[i]:y_to[i], x_from[i]:x_to[i]].copy()

    options = dict()
    options[DBOPT_CYCLE] = 99
    options[DBOPT_DTIME] = 0.99
    options[DBOPT_XLABEL] = "X"
    options[DBOPT_YLABEL] = "Y"
    options[DBOPT_XUNITS] = "a"
    options[DBOPT_YUNITS] = "b"
    options[DBOPT_HI_OFFSET] = hi_offset[i]

    mesh_name = "mesh"
    s.put_quadmesh(mesh_name, axes, coordtype=DB_COLLINEAR, optlist=options)
    mesh_names.append((f"{file_name}:{mesh_name}", DB_COLLINEAR))

    options = dict()
    options[DBOPT_UNITS] = "unit"

    var_name = "variable"
    s.put_quadvar1(var_name, mesh_name, data, data.shape,
            centering=DB_NODECENT, optlist=options)
    var_names.append((f"{file_name}:{mesh_name}", DBObjectType.DB_QUADVAR))

options = dict()
options[DBOPT_CYCLE] = 99
options[DBOPT_DTIME] = 0.99
options[DBOPT_XLABEL] = "xx"
options[DBOPT_YLABEL] = "yy"
options[DBOPT_XUNITS] = "a"
options[DBOPT_YUNITS] = "b"

s = SiloFile("example.silo",
        mode=DB_CLOBBER,
        filetype=DB_HDF5,
        target=DB_LOCAL,
        fileinfo="Example Metadata.")

s.put_multimesh("mesh", mesh_names, optlist=options)
s.put_multivar("scalar", var_names, optlist=options)



# modified from original code by Matthieu Haefele (IPP, Max-Planck-Gesellschaft)

import numpy
import pyvisfile.silo as silo

f = silo.SiloFile("qmesh.silo", mode=silo.DB_CLOBBER)
coord = [
        numpy.linspace(-1.0, 1.0, 50),
        numpy.linspace(-2.0, 2.0, 100)
        ]

f.put_quadmesh("meshxy", coord)

value = coord[0][:, numpy.newaxis] * coord[1][numpy.newaxis, :]

f.put_quadvar1("value", "meshxy", numpy.asarray(value, order="F"), value.shape,
        centering=silo.DB_NODECENT)

f.close()



import numpy as np
import pytest
from pyvisfile.vtk import (
    UnstructuredGrid, DataArray,
    AppendedDataXMLGenerator,
    VTK_VERTEX, VF_LIST_OF_VECTORS, VF_LIST_OF_COMPONENTS)

from pyvisfile.vtk import write_structured_grid


@pytest.mark.parametrize("n", [5000, 0])
def test_vtk_unstructured_points(n):
    points = np.random.randn(n, 3)

    data = [
            ("p", np.random.randn(n)),
            ("vel", np.random.randn(3, n)),
    ]
    file_name = f"points_{n}.vtu"
    compressor = None

    grid = UnstructuredGrid(
            (n, DataArray("points", points, vector_format=VF_LIST_OF_VECTORS)),
            cells=np.arange(n, dtype=np.uint32),
            cell_types=np.asarray([VTK_VERTEX] * n, dtype=np.uint8))

    for name, field in data:
        grid.add_pointdata(DataArray(name, field,
            vector_format=VF_LIST_OF_COMPONENTS))

    from os.path import exists
    if exists(file_name):
        raise RuntimeError("output file '%s' already exists"
            % file_name)

    outf = open(file_name, "w")
    AppendedDataXMLGenerator(compressor)(grid).write(outf)
    outf.close()


def test_vtk_structured_grid():
    angle_mesh = np.mgrid[1:2:10j, 0:2*np.pi:20j, 0:np.pi:30j]

    r = angle_mesh[0, np.newaxis]
    phi = angle_mesh[1, np.newaxis]
    theta = angle_mesh[2, np.newaxis]
    mesh = np.vstack((
        r*np.sin(theta)*np.cos(phi),
        r*np.sin(theta)*np.sin(phi),
        r*np.cos(theta),
        ))

    from pytools.obj_array import make_obj_array
    vec = make_obj_array([
        np.sin(theta)*np.cos(phi),
        np.sin(theta)*np.sin(phi),
        np.cos(theta),
        ])

    write_structured_grid("yo.vts", mesh,
            point_data=[("phi", phi), ("vec", vec)])


if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        exec(sys.argv[1])
    else:
        from pytest import main
        main([__file__])

# vim: fdm=marker



"""Generic support for new-style (XML) VTK visualization data files."""

__copyright__ = "Copyright (C) 2007 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import numpy as np

__doc__ = """

Constants
---------

Vector formats
^^^^^^^^^^^^^^

.. data:: VF_LIST_OF_COMPONENTS

    ``[[x0, y0, z0], [x1, y1, z1]]``

.. data:: VF_LIST_OF_VECTORS

    ``[[x0, x1], [y0, y1], [z0, z1]]``

Element types
^^^^^^^^^^^^^

.. data:: VTK_VERTEX
.. data:: VTK_POLY_VERTEX
.. data:: VTK_LINE
.. data:: VTK_POLY_LINE
.. data:: VTK_TRIANGLE
.. data:: VTK_TRIANGLE_STRIP
.. data:: VTK_POLYGON
.. data:: VTK_PIXEL
.. data:: VTK_QUAD
.. data:: VTK_TETRA
.. data:: VTK_VOXEL
.. data:: VTK_HEXAHEDRON
.. data:: VTK_WEDGE
.. data:: VTK_PYRAMID

.. data:: VTK_LAGRANGE_CURVE
.. data:: VTK_LAGRANGE_TRIANGLE
.. data:: VTK_LAGRANGE_QUADRILATERAL
.. data:: VTK_LAGRANGE_TETRAHEDRON
.. data:: VTK_LAGRANGE_HEXAHEDRON
.. data:: VTK_LAGRANGE_WEDGE

Building blocks
---------------

.. autoclass:: DataArray
.. autoclass:: UnstructuredGrid
.. autoclass:: StructuredGrid

XML elements
^^^^^^^^^^^^^^

.. autoclass:: XMLElement

XML generators
^^^^^^^^^^^^^^

.. autoclass:: InlineXMLGenerator
.. autoclass:: AppendedDataXMLGenerator
.. autoclass:: ParallelXMLGenerator

Convenience functions
---------------------

.. autofunction:: write_structured_grid
"""

# {{{ types

VTK_INT8 = "Int8"
VTK_UINT8 = "UInt8"
VTK_INT16 = "Int16"
VTK_UINT16 = "UInt16"
VTK_INT32 = "Int32"
VTK_UINT32 = "UInt32"
VTK_INT64 = "Int64"
VTK_UINT64 = "UInt64"
VTK_FLOAT32 = "Float32"
VTK_FLOAT64 = "Float64"


NUMPY_TO_VTK_TYPES = {
        np.int8: VTK_INT8,
        np.uint8: VTK_UINT8,
        np.int16: VTK_INT16,
        np.uint16: VTK_UINT16,
        np.int32: VTK_INT32,
        np.uint32: VTK_UINT32,
        np.int64: VTK_INT64,
        np.uint64: VTK_UINT64,
        np.float32: VTK_FLOAT32,
        np.float64: VTK_FLOAT64,
        }

# }}}


# {{{ cell types

# NOTE: should keep in sync with
# https://gitlab.kitware.com/vtk/vtk/-/blob/master/Common/DataModel/vtkCellType.h

# linear cells
VTK_VERTEX = 1
VTK_POLY_VERTEX = 2
VTK_LINE = 3
VTK_POLY_LINE = 4
VTK_TRIANGLE = 5
VTK_TRIANGLE_STRIP = 6
VTK_POLYGON = 7
VTK_PIXEL = 8
VTK_QUAD = 9
VTK_TETRA = 10
VTK_VOXEL = 11
VTK_HEXAHEDRON = 12
VTK_WEDGE = 13
VTK_PYRAMID = 14

# NOTE: these were added in VTK 8.1 as part of the commit
# https://gitlab.kitware.com/vtk/vtk/-/commit/cc5101a805386f205631357bba782b2a7d17531a

# high-order Lagrange cells
VTK_LAGRANGE_CURVE = 68
VTK_LAGRANGE_TRIANGLE = 69
VTK_LAGRANGE_QUADRILATERAL = 70
VTK_LAGRANGE_TETRAHEDRON = 71
VTK_LAGRANGE_HEXAHEDRON = 72
VTK_LAGRANGE_WEDGE = 73

# }}}


# {{{ cell node counts

CELL_NODE_COUNT = {
        VTK_VERTEX: 1,
        # VTK_POLY_VERTEX: no a-priori size
        VTK_LINE: 2,
        # VTK_POLY_LINE: no a-priori size
        VTK_TRIANGLE: 3,
        # VTK_TRIANGLE_STRIP: no a-priori size
        # VTK_POLYGON: no a-priori size
        VTK_PIXEL: 4,
        VTK_QUAD: 4,
        VTK_TETRA: 4,
        VTK_VOXEL: 8,
        VTK_HEXAHEDRON: 8,
        VTK_WEDGE: 6,
        VTK_PYRAMID: 5,
        # VTK_LAGRANGE_CURVE: no a-priori size
        # VTK_LAGRANGE_TRIANGLE: no a-priori size
        # VTK_LAGRANGE_QUADRILATERAL: no a-priori size
        # VTK_LAGRANGE_TETRAHEDRON: no a-priori size
        # VTK_LAGRANGE_HEXAHEDRON: no a-priori size
        # VTK_LAGRANGE_WEDGE: no a-priori size
        }

# }}}


# {{{ vector format

# e.g. [[x0, y0, z0], [x1, y1, z1]]
VF_LIST_OF_COMPONENTS = 0
# e.g. [[x0, x1], [y0, y1], [z0, z1]]
VF_LIST_OF_VECTORS = 1

# }}}


# {{{ xml

# Ah, the joys of home-baked non-compliant XML goodness.
class XMLElementBase:
    def __init__(self):
        self.children = []

    def copy(self, new_children=None):
        result = self.__class__(self.tag, self.attributes)
        if new_children is not None:
            result.children = new_children
        else:
            result.children = self.children
        return result

    def add_child(self, child):
        self.children.append(child)


class XMLElement(XMLElementBase):
    """
    .. automethod:: write
    """

    def __init__(self, tag, **attributes):
        XMLElementBase.__init__(self)
        self.tag = tag
        self.attributes = attributes

    def write(self, file):
        attr_string = "".join(
                f' {key}="{value}"'
                for key, value in self.attributes.items())
        if self.children:
            file.write(f"<{self.tag}{attr_string}>\n")
            for child in self.children:
                if isinstance(child, XMLElement):
                    child.write(file)
                else:
                    # likely a string instance, write it directly
                    file.write(child)
            file.write(f"</{self.tag}>\n")
        else:
            file.write(f"<{self.tag}{attr_string}/>\n")


class XMLRoot(XMLElementBase):
    def __init__(self, child=None):
        XMLElementBase.__init__(self)
        if child:
            self.add_child(child)

    def write(self, file):
        file.write('<?xml version="1.0"?>\n')
        for child in self.children:
            if isinstance(child, XMLElement):
                child.write(file)
            else:
                # likely a string instance, write it directly
                file.write(child)

# }}}


# {{{ encoded buffers

_U32CHAR = np.dtype(np.uint32).char


class EncodedBuffer:
    def encoder(self):
        """Return an identifier for the binary encoding used."""
        raise NotImplementedError

    def compressor(self):
        """Return an identifier for the compressor used, or None."""
        raise NotImplementedError

    def raw_buffer(self):
        """Reobtain the raw buffer string object that was used to
        construct this encoded buffer."""

        raise NotImplementedError

    def add_to_xml_element(self, xml_element):
        """Add encoded buffer to the given *xml_element*
        Return total size of encoded buffer in bytes."""

        raise NotImplementedError


class BinaryEncodedBuffer:
    def __init__(self, buffer):
        self.buffer = buffer

    def encoder(self):
        return "binary"

    def compressor(self):
        return None

    def raw_buffer(self):
        return self.buffer

    def add_to_xml_element(self, xml_element):
        raise NotImplementedError


class Base64EncodedBuffer:
    def __init__(self, buffer):
        from struct import pack
        from base64 import b64encode
        length = buffer.nbytes
        self.b64header = b64encode(
                pack(_U32CHAR, length)).decode()
        self.b64data = b64encode(buffer).decode()

    def encoder(self):
        return "base64"

    def compressor(self):
        return None

    def raw_buffer(self):
        from base64 import b64decode
        return b64decode(self.b64data)

    def add_to_xml_element(self, xml_element):
        """Add encoded buffer to the given *xml_element*.
        Return total size of encoded buffer in bytes."""

        xml_element.add_child(self.b64header)
        xml_element.add_child(self.b64data)

        return len(self.b64header) + len(self.b64data)


class Base64ZLibEncodedBuffer:
    def __init__(self, buffer):
        from struct import pack
        from base64 import b64encode
        from zlib import compress
        comp_buffer = compress(buffer)
        comp_header = [1, len(buffer), len(buffer), len(comp_buffer)]
        self.b64header = b64encode(
                pack(_U32CHAR*len(comp_header), *comp_header))
        self.b64data = b64encode(comp_buffer)

    def encoder(self):
        return "base64"

    def compressor(self):
        return "zlib"

    def raw_buffer(self):
        from base64 import b64decode
        from zlib import decompress
        return decompress(b64decode(self.b64data))

    def add_to_xml_element(self, xml_element):
        """Add encoded buffer to the given *xml_element*.
        Return total size of encoded buffer in bytes."""

        xml_element.add_child(self.b64header)
        xml_element.add_child(self.b64data)

        return len(self.b64header) + len(self.b64data)

# }}}


# {{{ data array

class DataArray:
    def __init__(self, name, container, vector_padding=3,
            vector_format=VF_LIST_OF_COMPONENTS, components=None):
        self.name = name

        if isinstance(container, DataArray):
            self.type = container.type
            self.components = container.components
            self.encoded_buffer = container.encoded_buffer
            return

        if not isinstance(container, np.ndarray):
            raise ValueError(
                f"cannot convert object of type '{type(container)}' to DataArray")

        if container.dtype.char == "O":
            for subvec in container:
                if not isinstance(subvec, np.ndarray):
                    raise TypeError(
                            f"expected numpy array, got '{type(subvec)}' instead")

            container = np.array(list(container))
            assert container.dtype.char != "O"

        if len(container.shape) > 1:
            if vector_format == VF_LIST_OF_COMPONENTS:
                container = container.T.copy()

            assert len(container.shape) == 2, \
                    "numpy vectors of rank >2 are not supported"
            assert container.strides[1] == container.itemsize, \
                    "2D numpy arrays must be row-major"

            if vector_padding > container.shape[1]:
                container = np.asarray(np.hstack((
                        container,
                        np.zeros((
                            container.shape[0],
                            vector_padding-container.shape[1],
                            ),
                            container.dtype))), order="C")
            self.components = container.shape[1]
        else:
            self.components = 1

        self.type = NUMPY_TO_VTK_TYPES.get(container.dtype.type, None)
        if self.type is None:
            raise TypeError(f"unsupported vector type: '{container.dtype}'")

        if not container.flags.c_contiguous:
            container = container.copy()

        buf = memoryview(container)
        self.encoded_buffer = BinaryEncodedBuffer(buf)

    def get_encoded_buffer(self, encoder, compressor):
        have_encoder = self.encoded_buffer.encoder()
        have_compressor = self.encoded_buffer.compressor()

        if (encoder, compressor) != (have_encoder, have_compressor):
            raw_buf = self.encoded_buffer.raw_buffer()

            # avoid having three copies of the buffer around temporarily
            del self.encoded_buffer

            if (encoder, compressor) == ("binary", None):
                self.encoded_buffer = BinaryEncodedBuffer(raw_buf)
            elif (encoder, compressor) == ("base64", None):
                self.encoded_buffer = Base64EncodedBuffer(raw_buf)
            elif (encoder, compressor) == ("base64", "zlib"):
                self.encoded_buffer = Base64ZLibEncodedBuffer(raw_buf)
            else:
                self.encoded_buffer = BinaryEncodedBuffer(raw_buf)
                raise ValueError("invalid encoder/compressor pair")

            have_encoder = self.encoded_buffer.encoder()
            have_compressor = self.encoded_buffer.compressor()

            assert (encoder, compressor) == (have_encoder, have_compressor)

        return self.encoded_buffer

    def encode(self, compressor, xml_element):
        ebuf = self.get_encoded_buffer("base64", compressor)
        return ebuf.add_to_xml_element(xml_element)

    def invoke_visitor(self, visitor):
        return visitor.gen_data_array(self)

# }}}


# {{{ grids

class UnstructuredGrid:
    """
    .. automethod:: add_pointdata
    .. automethod:: add_celldata
    """

    def __init__(self, points, cells, cell_types):
        self.cell_count = len(cells)

        self.point_count, self.points = points
        assert self.points.name == "points"

        try:
            self.cell_count, self.cell_connectivity, self.cell_offsets = cells
        except Exception:
            self.cell_count = len(cell_types)

            if self.cell_count > 0:
                offsets = np.cumsum(
                        np.vectorize(CELL_NODE_COUNT.get)(cell_types),
                        dtype=cells.dtype)
            else:
                offsets = np.empty((0,), dtype=cells.dtype)

            self.cell_connectivity = DataArray("connectivity", cells)
            self.cell_offsets = DataArray("offsets", offsets)

        self.cell_types = DataArray("types", cell_types)

        self.pointdata = []
        self.celldata = []

    def copy(self):
        return UnstructuredGrid(
                (self.point_count, self.points),
                (self.cell_count, self.cell_connectivity, self.cell_offsets),
                self.cell_types)

    def vtk_extension(self):
        return "vtu"

    def invoke_visitor(self, visitor):
        return visitor.gen_unstructured_grid(self)

    def add_pointdata(self, data_array):
        self.pointdata.append(data_array)

    def add_celldata(self, data_array):
        self.celldata.append(data_array)


class StructuredGrid:
    """
    .. automethod:: add_pointdata
    .. automethod:: add_celldata
    """

    def __init__(self, mesh):
        """
        :arg mesh: has shape *(ndims, nx, ny, nz)*
            (ny, nz may be omitted)
        """
        self.mesh = mesh

        self.ndims = mesh.shape[0]
        transpose_arg = tuple(range(1, 1+self.ndims)) + (0,)
        mesh = mesh.transpose(transpose_arg).copy()

        self.shape = mesh.shape[:-1][::-1]
        mesh = mesh.reshape(-1, self.ndims)
        self.points = DataArray(
                "points", mesh,
                vector_format=VF_LIST_OF_VECTORS)

        self.pointdata = []
        self.celldata = []

    def copy(self):
        return StructuredGrid(self.mesh)

    def vtk_extension(self):
        return "vts"

    def invoke_visitor(self, visitor):
        return visitor.gen_structured_grid(self)

    def add_pointdata(self, data_array):
        self.pointdata.append(data_array)

    def add_celldata(self, data_array):
        self.celldata.append(data_array)

# }}}


# {{{ vtk xml writers

def make_vtkfile(filetype, compressor, version="0.1"):
    import sys
    if sys.byteorder == "little":
        bo = "LittleEndian"
    else:
        bo = "BigEndian"

    kwargs = {}
    if compressor == "zlib":
        kwargs["compressor"] = "vtkZLibDataCompressor"

    return XMLElement("VTKFile",
            type=filetype, version=version, byte_order=bo, **kwargs)


class XMLGenerator:
    def __init__(self, compressor=None, vtk_file_version=None):
        """
        :arg vtk_file_version: a string ``"x.y"`` with the desired VTK
            XML file format version. Relevant versions are as follows:

            * ``"0.1"`` is the original version.
            * ``"1.0"`` added support for 64-bit indices and offsets, as
              described `here <https://www.paraview.org/Wiki/VTK_XML_Formats>`__.
            * ``"2.0"`` added support for ghost array data, as
              described `here <https://blog.kitware.com/ghost-and-blanking-visibility-changes/>`__.
            * ``"2.1"``: added support for writing additional information
              attached to a :class:`DataArray` using
              `information keys <https://vtk.org/doc/nightly/html/IOXMLInformationFormat.html>`__.
            * ``"2.2"``: changed the node numbering of the hexahedron, as
              described `here <https://gitlab.kitware.com/vtk/vtk/-/merge_requests/6678>`__.
        """     # noqa

        if compressor == "zlib":
            try:
                import zlib  # noqa
            except ImportError:
                compressor = None
        elif compressor is None:
            pass
        else:
            raise ValueError(f"invalid compressor name '{compressor}'")

        if vtk_file_version is None:
            # https://www.paraview.org/Wiki/VTK_XML_Formats
            vtk_file_version = "0.1"

        self.vtk_file_version = vtk_file_version
        self.compressor = compressor

    def __call__(self, vtkobj):
        """Return an :class:`XMLElement`."""

        child = self.rec(vtkobj)
        vtkf = make_vtkfile(child.tag, self.compressor,
                version=self.vtk_file_version)
        vtkf.add_child(child)
        return XMLRoot(vtkf)

    def rec(self, vtkobj):
        return vtkobj.invoke_visitor(self)


class InlineXMLGenerator(XMLGenerator):
    """
    .. automethod:: __init__
    .. automethod:: __call__
    """

    def gen_unstructured_grid(self, ugrid):
        el = XMLElement("UnstructuredGrid")
        piece = XMLElement("Piece",
                NumberOfPoints=ugrid.point_count, NumberOfCells=ugrid.cell_count)
        el.add_child(piece)

        if ugrid.pointdata:
            data_el = XMLElement("PointData")
            piece.add_child(data_el)
            for data_array in ugrid.pointdata:
                data_el.add_child(self.rec(data_array))

        if ugrid.celldata:
            data_el = XMLElement("CellData")
            piece.add_child(data_el)
            for data_array in ugrid.celldata:
                data_el.add_child(self.rec(data_array))

        points = XMLElement("Points")
        piece.add_child(points)
        points.add_child(self.rec(ugrid.points))

        cells = XMLElement("Cells")
        piece.add_child(cells)
        cells.add_child(self.rec(ugrid.cell_connectivity))
        cells.add_child(self.rec(ugrid.cell_offsets))
        cells.add_child(self.rec(ugrid.cell_types))

        return el

    def gen_structured_grid(self, sgrid):
        extent = []
        for dim in range(3):
            extent.append(0)
            if dim < sgrid.ndims:
                extent.append(sgrid.shape[dim]-1)
            else:
                extent.append(0)
        extent_str = " ".join(str(i) for i in extent)

        el = XMLElement("StructuredGrid", WholeExtent=extent_str)
        piece = XMLElement("Piece", Extent=extent_str)
        el.add_child(piece)

        if sgrid.pointdata:
            data_el = XMLElement("PointData")
            piece.add_child(data_el)
            for data_array in sgrid.pointdata:
                data_el.add_child(self.rec(data_array))

        if sgrid.celldata:
            data_el = XMLElement("CellData")
            piece.add_child(data_el)
            for data_array in sgrid.celldata:
                data_el.add_child(self.rec(data_array))

        points = XMLElement("Points")
        piece.add_child(points)
        points.add_child(self.rec(sgrid.points))
        return el

    def gen_data_array(self, data):
        el = XMLElement("DataArray", type=data.type, Name=data.name,
                NumberOfComponents=data.components, format="binary")
        data.encode(self.compressor, el)
        el.add_child("\n")
        return el


class AppendedDataXMLGenerator(InlineXMLGenerator):
    """
    .. automethod:: __call__
    """

    def __init__(self, compressor=None, vtk_file_version=None):
        InlineXMLGenerator.__init__(self, compressor, vtk_file_version)

        self.base64_len = 0
        self.app_data = XMLElement("AppendedData", encoding="base64")
        self.app_data.add_child("_")

    def __call__(self, vtkobj):
        """Return an :class:`XMLElement`."""

        xmlroot = XMLGenerator.__call__(self, vtkobj)
        self.app_data.add_child("\n")
        xmlroot.children[0].add_child(self.app_data)
        return xmlroot

    def gen_data_array(self, data):
        el = XMLElement("DataArray", type=data.type, Name=data.name,
                NumberOfComponents=data.components, format="appended",
                offset=self.base64_len)

        self.base64_len += data.encode(self.compressor, self.app_data)

        return el


class ParallelXMLGenerator(XMLGenerator):
    """
    .. automethod:: __call__
    """

    def __init__(self, pathnames):
        XMLGenerator.__init__(self, compressor=None)

        self.pathnames = pathnames

    def gen_unstructured_grid(self, ugrid):
        el = XMLElement("PUnstructuredGrid")

        pointdata = XMLElement("PPointData")
        el.add_child(pointdata)
        for data_array in ugrid.pointdata:
            pointdata.add_child(self.rec(data_array))

        points = XMLElement("PPoints")
        el.add_child(points)
        points.add_child(self.rec(ugrid.points))

        cells = XMLElement("PCells")
        el.add_child(cells)
        cells.add_child(self.rec(ugrid.cell_connectivity))
        cells.add_child(self.rec(ugrid.cell_offsets))
        cells.add_child(self.rec(ugrid.cell_types))

        for pn in self.pathnames:
            el.add_child(XMLElement("Piece", Source=pn))

        return el

    def gen_data_array(self, data):
        el = XMLElement("PDataArray", type=data.type, Name=data.name,
                NumberOfComponents=data.components)
        return el


def write_structured_grid(
        file_name, mesh,
        cell_data=None, point_data=None,
        overwrite=False):
    if cell_data is None:
        cell_data = []

    if point_data is None:
        point_data = []

    grid = StructuredGrid(mesh)

    from pytools.obj_array import obj_array_vectorize

    def do_reshape(fld):
        return fld.T.copy().reshape(-1)

    for name, field in cell_data:
        reshaped_fld = obj_array_vectorize(do_reshape, field)
        grid.add_pointdata(DataArray(name, reshaped_fld))

    for name, field in point_data:
        reshaped_fld = obj_array_vectorize(do_reshape, field)
        grid.add_pointdata(DataArray(name, reshaped_fld))

    import os.path
    if os.path.exists(file_name):
        if overwrite:
            # nothing to do, just overwrite below
            pass
        else:
            raise FileExistsError(f"output file '{file_name}' already exists")

    with open(file_name, "w") as outf:
        AppendedDataXMLGenerator()(grid).write(outf)



from urllib.request import urlopen

_conf_url = \
        "https://raw.githubusercontent.com/inducer/sphinxconfig/main/sphinxconfig.py"
with urlopen(_conf_url) as _inf:
    exec(compile(_inf.read(), _conf_url, "exec"), globals())

copyright = "2010-21, Andreas Kloeckner"

ver_dic = {}
with open("../pyvisfile/__init__.py") as vpy_file:
    version_py = vpy_file.read()

exec(compile(version_py, "../pyvisfile/__init__.py", "exec"), ver_dic)
version = ".".join(str(x) for x in ver_dic["VERSION"])
# The full version, including alpha/beta/rc tags.
release = version

intersphinx_mapping = {
    "https://docs.python.org/3/": None,
    "https://numpy.org/doc/stable/": None,
    "https://documen.tician.de/modepy": None,
}



#!/usr/bin/env python

from aksetup_helper import NumpyExtension


class PyUblasExtension(NumpyExtension):
    def get_module_include_path(self, name):
        from pkg_resources import Requirement, resource_filename
        return resource_filename(Requirement.parse(name), "%s/include" % name)

    def get_additional_include_dirs(self):
        return (NumpyExtension.get_additional_include_dirs(self)
                + [self.get_module_include_path("pyublas")])


def get_config_schema():
    from aksetup_helper import (ConfigSchema,
            IncludeDir, LibraryDir, Libraries,
            Switch, StringListOption, make_boost_base_options)

    return ConfigSchema(make_boost_base_options() + [
        Switch("USE_SILO", False, "Compile libsilo interface"),

        IncludeDir("SILO", []),
        LibraryDir("SILO", []),
        Libraries("SILO", ["siloh5"]),

        StringListOption("CXXFLAGS", [],
            help="Any extra C++ compiler options to include"),
        ])


def main():
    from setuptools import find_packages
    from aksetup_helper import (hack_distutils, get_config, setup,
            ExtensionUsingNumpy, check_pybind11, PybindBuildExtCommand,
            get_pybind_include)

    hack_distutils()
    conf = get_config(get_config_schema(),
            warn_about_no_config=False)

    extra_defines = {}
    extra_include_dirs = []
    extra_library_dirs = []
    extra_libraries = []

    ver_dic = {}
    ver_file_name = "pyvisfile/__init__.py"
    with open(ver_file_name) as inf:
        exec(compile(inf.read(), ver_file_name, "exec"), ver_dic)

    requirements = []
    ext_modules = []

    if conf["USE_SILO"]:
        check_pybind11()

        extra_defines["USE_SILO"] = 1
        extra_include_dirs.extend(conf["SILO_INC_DIR"])
        extra_library_dirs.extend(conf["SILO_LIB_DIR"])
        extra_libraries.extend(conf["SILO_LIBNAME"])

        ext_modules.append(ExtensionUsingNumpy("_internal",
            ["src/wrapper/wrap_silo.cpp"],
            include_dirs=[get_pybind_include()] + extra_include_dirs,
            library_dirs=extra_library_dirs,
            libraries=extra_libraries,
            extra_compile_args=conf["CXXFLAGS"],
            define_macros=list(extra_defines.items()),
            language="c++",
            ))
        requirements.append("pybind11>=2.5.0")

    setup(name="pyvisfile",
            version=ver_dic["VERSION_TEXT"],
            description="Large-scale Visualization Data Storage",
            long_description=open("README.rst").read(),
            classifiers=[
                "Development Status :: 4 - Beta",
                "Intended Audience :: Developers",
                "Intended Audience :: Other Audience",
                "Intended Audience :: Science/Research",
                "License :: OSI Approved :: MIT License",
                "Natural Language :: English",
                "Programming Language :: C++",
                "Programming Language :: Python",
                "Programming Language :: Python :: 3",
                "Topic :: Multimedia :: Graphics :: 3D Modeling",
                "Topic :: Scientific/Engineering",
                "Topic :: Scientific/Engineering :: Mathematics",
                "Topic :: Scientific/Engineering :: Physics",
                "Topic :: Scientific/Engineering :: Visualization",
                "Topic :: Software Development :: Libraries",
                ],

            author="Andreas Kloeckner",
            author_email="inform@tiker.net",
            license="MIT",
            url="http://mathema.tician.de/software/pyvisfile",

            # dependencies
            setup_requires=requirements,
            python_requires="~=3.6",
            install_requires=[
                    "pytools>=2013.2",
                    ] + requirements,

            packages=find_packages(),
            ext_package="pyvisfile.silo",
            ext_modules=ext_modules,

            cmdclass={"build_ext": PybindBuildExtCommand},
            zip_safe=False)


if __name__ == "__main__":
    main()
